/*
 * central command for setting up several tasks
 * 
 * rf, 6/21/94
 *
 * rf, 4/13/99
 *  rename hqp.tcl to hqp_solve.tcl for initial evaluation
 *
 * rf, 12/25/99
 *  add support for Tcl package handling
 *
 */

/*
    Copyright (C) 1994--1999  Ruediger Franke

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; 
    version 2 of the License.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library (file COPYING.LIB);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <signal.h>
#include <tcl.h>

#include <If_List.h>
#include <If_Element.h>
#include <If_Proc.h>
#include <If_Class.h>
#include <If_Module.h>

#include "Hqp_SqpSolver.h"
#include "Hqp_SqpProgram.h"

#include "Hqp_SqpPowell.h"

// define a global Tcl interpreter for interface elements
//-------------------------------------------------------
Tcl_Interp *theInterp = NULL;

static If_List _ifList;

Hqp_SqpProgram *theSqpProgram;
Hqp_SqpSolver *theSqpSolver;

#ifdef IF_CLASS_STATIC
//--------------------------------------------------------------------------
// ensure linkage of modules
//--------------------------------------------------------------------------
#if defined(HQP_WITH_ASCEND)
#include "Prg_ASCEND.h"
#endif
#include "Prg_CUTE.h"
#include "Prg_CUTE_ST.h"
#include "Hqp_SqpSchittkowski.h"
#include "Hqp_HL_BFGS.h"
#include "Hqp_HL_DScale.h"
#include "Hqp_HL_Gerschgorin.h"
#include "Hqp_IpRedSpBKP.h"
#include "Hqp_IpSpBKP.h"
#include "Hqp_IpLQDOCP.h"
#include "Hqp_IpSpSC.h"
#include "Hqp_IpsFranke.h"
#include "Hqp_IpsMehrotra.h"
#include "Hqp_Client.h"

static void Hqp_ClassAlloc()
{
  IF_CLASS_ALLOC("Client", Hqp_Client, Hqp_Solver);
  IF_CLASS_ALLOC("Mehrotra", Hqp_IpsMehrotra, Hqp_Solver);
  IF_CLASS_ALLOC("Franke", Hqp_IpsFranke, Hqp_Solver);
  IF_CLASS_ALLOC("SpSC", Hqp_IpSpSC, Hqp_IpMatrix);
  IF_CLASS_ALLOC("LQDOCP", Hqp_IpLQDOCP, Hqp_IpMatrix);
  IF_CLASS_ALLOC("SpBKP", Hqp_IpSpBKP, Hqp_IpMatrix);
  IF_CLASS_ALLOC("RedSpBKP", Hqp_IpRedSpBKP, Hqp_IpMatrix);
  IF_CLASS_ALLOC("Gerschgorin", Hqp_HL_Gerschgorin, Hqp_HL);
  IF_CLASS_ALLOC("DScale", Hqp_HL_DScale, Hqp_HL);
  IF_CLASS_ALLOC("BFGS", Hqp_HL_BFGS, Hqp_HL);
  IF_CLASS_ALLOC("Schittkowski", Hqp_SqpSchittkowski, Hqp_SqpSolver);
  IF_CLASS_ALLOC("Powell", Hqp_SqpPowell, Hqp_SqpSolver);
  IF_CLASS_ALLOC("CUTE_ST", Prg_CUTE_ST, Hqp_SqpProgram);
  IF_CLASS_ALLOC("CUTE", Prg_CUTE, Hqp_SqpProgram);
#if defined(HQP_WITH_ASCEND)
  IF_CLASS_ALLOC("ASCEND", Prg_ASCEND, Hqp_SqpProgram);
#endif
}
#endif

//--------------------------------------------------------------------------
// return HQP's version
//--------------------------------------------------------------------------
const char *Hqp_Version = VERSION;

static int Hqp_VersionCmd(int, char *[], char **result)
{
  *result = (char *)Hqp_Version;
  return IF_OK;
}

//--------------------------------------------------------------------------
// fulfill the Meschach copyright
//--------------------------------------------------------------------------
static int m_version_cmd(int, char *[], char **)
{
  m_version();

  return IF_OK;
}

//--------------------------------------------------------------------------
static void signal_handler(int code)
{
  switch (code) {

  case SIGINT:
    Tcl_Eval(theInterp, "hqp_exit {signal interrupt}");
    fprintf(stderr, "HQP %s: %s\n", Hqp_Version, "signal interrupt");
    break;

  case SIGXCPU:
    signal(SIGXCPU, SIG_IGN);
    Tcl_Eval(theInterp, "hqp_exit {signal cputime}");
    fprintf(stderr, "HQP %s: %s\n", Hqp_Version, "signal cputime");
    break;

  default:
    break;
  }

  fprintf(stderr, "%s\n", theInterp->result);
  exit(0);
}

//--------------------------------------------------------------------------
extern "C" int Hqp_Init(Tcl_Interp *interp) 
{
  // provide Tcl package Hqp
  if (Tcl_PkgRequire(interp, "Tcl", "8.0", 0) == NULL ||
      Tcl_PkgProvide(interp, "Hqp", (char *)Hqp_Version) != TCL_OK) {
    return TCL_ERROR;
  }
#if 0 
  // don't use error handling as this doesn't seem to work
  // for Hqp a shared lib and loaded to an executable
  // rf, 2/7/97
  int code;
  char *reason;

  if ((code = setjmp(restart)) != 0) {

    set_err_flag(EF_EXIT);	// avoid recursive error calls

    switch (code) {
    case E_MEM:
      reason = "out of memory";
      break;
    case E_SING:
      reason = "singular matrix";
      break;
    default:
      reason = "execution error";
    }

    Tcl_VarEval(theInterp, "hqp_exit {", reason, "}", NULL);
    fprintf(stderr, "HQP %s: %s\n", Hqp_Version, reason);
    fprintf(stderr, "%s\n", theInterp->result);
    exit(-1);
  }
  set_err_flag(EF_JUMP);
#endif

  // export interpreter
  theInterp = interp;

  // allocate interface modules
# ifdef IF_CLASS_STATIC
  Hqp_ClassAlloc();
# endif

  // create initial state
  theSqpProgram = NULL;
  theSqpSolver = new Hqp_SqpPowell;
  _ifList.append(new IF_MODULE("sqp_solver", &theSqpSolver, Hqp_SqpSolver));
  _ifList.append(new IF_MODULE("prg_name", &theSqpProgram, Hqp_SqpProgram));
  _ifList.append(new If_Proc("hqp_version", &Hqp_VersionCmd));
  _ifList.append(new If_Proc("m_version", &m_version_cmd));

  // install a handler for signal interrupt
  signal(SIGINT, &signal_handler);
  signal(SIGXCPU, &signal_handler);

  // ignore signals from floating point arithmetics
  // (needed for ADOL-C 1.7 on Alpha with OSF3.2, OSF4.0)
  signal(SIGFPE, SIG_IGN);

  // evaluate hqp_solve.tcl
  extern char *hqp_solve;
  if (Tcl_Eval(interp, hqp_solve) == TCL_ERROR) {
    fprintf(stderr,
	    "Execution of built-in code failed: %s\n", interp->result);
  }

  return TCL_OK;
}
